/*
 *	dial -- dial the phone number
 *
 * Copyright 1989 by W J Andrew Fyfe.  All rights reserved.  Permission to
 * copy this program is given provided that the copy is not sold and that
 * this copyright notice is included.
 */

#include "uw.h"

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <dial.h>
#include <sys/phone.h>

static const char * const errs[] = { "zero", "interrupt", "dialer hung",
			      "no answer", "illegal baud-rate", "acu problem",
			      "line problem", "can't open LDEVS",
			      "device not available", "device unknown",
			      "no device available at baud-rate",
			      "no device known at baud-rate" };

int dial(CALL);
void undial(int);
static int translate_baud(const char *speed);
static int is_phone = 0;

int
dial_number(const char *line, const char *speed, const char *number)
{
    CALL call;
    struct termio term;
    struct updata phone;
    int fd;
    int baud;

    fd = open(line, O_RDWR | O_NDELAY, 0);
    if (fd < 0) {
	fprintf(stderr, "%s: Can't open %s: ", cmd_name, line);
	perror("");
	return -1;
    }
    if (strncmp(line, PHDEV, sizeof(PHDEV)-1) == 0) {
	is_phone = 1;
	if (ioctl(fd, PIOCGETP, &phone) >= 0) {
	    phone.c_lineparam &= ~PULSE;
	    phone.c_lineparam |= DTMF;
	    phone.c_feedback &= ~(SOFTSPK | NORMSPK | LOUDSPK);
	    phone.c_feedback |= (SPEAKERON | SOFTSPK);
	    ioctl(fd, PIOCSETP, &phone);
	}
    }
    ioctl(fd, TCGETA, &term);
    close(fd);

    baud = translate_baud(speed);
    if (baud < 0) {
	fprintf(stderr, "%s: illegal speed: \"%s\"\n", cmd_name, speed);
	return -1;
    }

    term.c_iflag = IXOFF | IXON | ISTRIP | IGNPAR;
    term.c_oflag = 0;
    term.c_cflag = HUPCL | CREAD | CS8 | baud;
    term.c_lflag = 0;
    term.c_cc[VMIN] = 1;
    term.c_cc[VTIME] = 0;
    call.attr = &term;
    call.baud = call.speed = atoi(speed);
    call.line = strrchr(line, '/') + 1;
    call.telno = (char *)number;
    call.modem = 0;

    fd = dial(call);
    if (fd < 0) {
	fprintf(stderr, "%s: dial failed: %s\n", cmd_name, errs[-fd]);
	return -1;
    }

    return fd;
}

void
undial_number(const char *line)
{
    if (phone_fd >= 0) {
	if (is_phone)
	    ioctl(phone_fd, PIOCDISC, (char *)0);
	undial(phone_fd);
	close(phone_fd);
    }
    
}

static const struct trans {
    char *speed;
    int baud;
} trans[] = {
	{"0",     0},
	{"50",    B50},
	{"75",    B75},
	{"110",   B110},
	{"134",   B134},
	{"150",   B150},
	{"200",   B200},
	{"300",   B300},
	{"600",   B600},
	{"1200",  B1200},
	{"1800",  B1800},
	{"2400",  B2400},
	{"4800",  B4800},
	{"9600",  B9600},
	{"19200", B19200},
};

static int
translate_baud(const char *speed)
{
     int i;

     for (i=0; i < sizeof(trans)/sizeof(struct trans); ++i) {
	 if (strcmp(speed, trans[i].speed) == 0)
	     return trans[i].baud;
     }
     return -1;
}

int
modem_connected(void)
{
    struct updata d;

    if (!is_phone || ioctl(phone_fd, PIOCGETP, &d) < 0)
	return 1;

    return d.c_linestatus & MODEMCONNECTED;
}
