/*
 *	reader -- get input from the keyboard and sent it to the host
 *
 * Copyright 1989 by W J Andrew Fyfe.  All rights reserved.  Permission to
 * copy this program is given provided that the copy is not sold and that
 * this copyright notice is included.
 */
 
#include "uw.h"
#include <sys/window.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <termio.h>
#include <signal.h>
#include <string.h>

static int set_current(int real_window);
static void check_input(void);
static void check_command(void);
static void send_to_server(unsigned char);
static void prompt_user(int fd, const char *prompt, char input[WTXTLEN]);

static int window = -1, sig_usr1 = 1, sig_alrm;

static void
catch_usr1(void)
{
    ++sig_usr1;
    signal(SIGUSR1, catch_usr1);
}

static void
catch_wind(void)
{
    signal(SIGWIND, catch_wind);
}

static void
catch_alrm(void)
{
    sig_alrm = 1;
    signal(SIGALRM, catch_alrm);
    alarm(TIMEOUT);
}

void
reader(void)
{
    int current;

    signal(SIGUSR1, catch_usr1);
    signal(SIGWIND, catch_wind);
    signal(SIGALRM, catch_alrm);
    alarm(TIMEOUT);

    while (1) {
	if (sig_usr1 || sig_alrm)
	    check_command();

	current = ioctl(STDIN, WIOCGCURR, (char *)0);
	if (current >= 0 && (window < 0 || 
	    (window >= 0 && current != win_info[window].window_number)))
	    if (set_current(current) < 0 && !sig_usr1)
		pause();

	if (window >= 0)
	    check_input();
    }
}

static int
set_current(int real_window)
{
    int i;

    if (protocol == PROTOCOL_0) {
	if (win_info[0].window_number == real_window && win_info[0].fd >= 0) {
	    window = 0;
	    return 0;
	}
    }
    else if (protocol == PROTOCOL_1) {
	for (i=1; i<=MAX_WINDOWS; ++i) {
	    if (win_info[i].window_number == real_window
		&& win_info[i].fd >= 0) {
		send_to_server(P1_FN_ISELW | i);
		window = i;
		return 0;
	    }
	}
    }

    return -1;
}

static int text_line = -1;

static void
check_input(void)
{
    int i;
    struct utdata t;
    static int keyboard_escape = 0;
    unsigned char c;
    struct message msg;
    char buffer[WTXTLEN];
    static const char *text_message[] = {
	"send <B>reak"
	    "    <C>hange directory"
	    "    <D>ownload file"
	    "    <U>pload file"
	    "    e<X>it uw",
	"<C>reate window"
	    "    <K>ill window"
	    "    Select window <#######>"
    };

    if (read(win_info[window].fd, &c, 1) <= 0)
	return;

    if (!keyboard_escape) {
	if (c == kbd_esc_char) {
	    keyboard_escape = 1;
	    text_line = window;
	    t.ut_num = WTXTSLK1;
	    if (protocol == PROTOCOL_0)
		strcpy(t.ut_text, text_message[0]);
	    if (protocol == PROTOCOL_1) {
		int i = strchr(text_message[1], '#') - text_message[1];
		int j;

		strcpy(t.ut_text, text_message[1]);
		for (j=1; j<=7; ++j, ++i)
		    t.ut_text[i] = (win_info[j].fd >= 0) ? j+'0' : ' ';
	    }
	    ioctl(win_info[text_line].fd, WIOCSETTEXT, &t);
	    return;
	}
	if (use_meta && (c & META_BIT)) {
	    keyboard_escape = 1;
	    c &= ~META_BIT;
	}
    }
    if (protocol == PROTOCOL_0) {
	if (text_line >= 0) {
	    t.ut_num = WTXTSLK1;
	    memset(t.ut_text, ' ', WTXTLEN-1);
	    t.ut_text[WTXTLEN-1] = '\0';
	    ioctl(win_info[text_line].fd, WIOCSETTEXT, &t);
	    text_line = -1;
	}
	if (!keyboard_escape || (keyboard_escape && c == kbd_esc_char)) {
	    write(phone_fd, &c, 1);
	}
	else {
	    switch(c) {
	    case 'b':
	    case 'B':
		ioctl(phone_fd, TCSBRK, (char *)0);
		break;
#if 0
	    case 's':
	    case 'S':
		send_controller(P1_INT_STARTUP);
		break;
#endif
	    case 'c':
	    case 'C':
		buffer[0] = '\0';
		prompt_user(win_info[window].fd, "Change Directory", buffer);
		send_controller_string(P1_INT_CHDIR, buffer);
		while (msgrcv(r_id, &msg, MSG_SIZE, MSG_TYPE, 0) < 0)
		    pause();
		text_line = window;
		t.ut_num = WTXTSLK1;
		strncpy(t.ut_text, &msg.text[STRING_VALUE], WTXTLEN-1);
		t.ut_text[WTXTLEN-1] = '\0';
		ioctl(win_info[text_line].fd, WIOCSETTEXT, &t);
		break;
	    case 'u':
	    case 'U':
		buffer[0] = '\0';
		prompt_user(win_info[window].fd, "Upload File Name", buffer);
		if (buffer[0] != '\0')
		    send_controller_string(P1_INT_UPLOAD, buffer);
		break;
	    case 'd':
	    case 'D':
		send_controller(P1_INT_DOWNLOAD);
		break;
	    case 'x':
	    case 'X':
		send_controller(P1_INT_DIE);
		break;
	    default:
		c = '\a';
		write(1, &c, 1);
		break;
	    }
	}
    }
    if (protocol == PROTOCOL_1) {
	if (text_line >= 0) {
	    t.ut_num = WTXTSLK1;
	    memset(t.ut_text, ' ', WTXTLEN-1);
	    t.ut_text[WTXTLEN-1] = '\0';
	    ioctl(win_info[text_line].fd, WIOCSETTEXT, &t);
	    text_line = -1;
	}
	if (!keyboard_escape || (keyboard_escape && c == kbd_esc_char)) {
	    if (c & META_BIT) {
		send_to_server(P1_FN_META);
		c &= ~META_BIT;
	    }
	    switch(c) {
	    case IAC:
		send_to_server(P1_FN_CTLCH | P1_CC_IAC);
		break;
	    case XON:
		send_to_server(P1_FN_CTLCH | P1_CC_XON);
		break;
	    case XOFF:
		send_to_server(P1_FN_CTLCH | P1_CC_XOFF);
		break;
	    default:
		write(phone_fd, &c, 1);
		break;
	    }
	}
	else {
	    switch(c) {
#if 0
	    case 's':
	    case 'S':
		send_controller(P1_INT_SHUTDOWN);
		break;
#endif
	    case 'C':
	    case 'c':
		for (i=1; i<=MAX_WINDOWS; ++i) {
		    if (win_info[i].fd < 0) {
			send_controller(P1_FN_NEWW | i);
			send_to_server(P1_FN_NEWW | i);
			break;
		    }
		}
		break;
	    case 'K':
	    case 'k':
		if (window > 0 && win_info[window].fd >= 0) {
		    send_controller(P1_FN_KILLW | window);
		    send_to_server(P1_FN_KILLW | window);
		}
		break;
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
		i = c - '1' + 1;
		if (win_info[i].fd >= 0)
		    ioctl(win_info[i].fd, WIOCSELECT, (char *)0);
		break;
	    default:
		c = '\a';
		write(1, &c, 1);
		break;
	    }
	}
    }
    keyboard_escape = 0;
    return;
}

static void
check_command(void)
{
    struct message msg;
    struct utdata t;

    sig_alrm = 0;

    if (msgrcv(r_id, &msg, MSG_SIZE, MSG_TYPE, IPC_NOWAIT) < 0)
	return;

    if (sig_usr1 > 0)
	--sig_usr1;

    switch (msg.text[COMMAND]) {
    case PROTOCOL_0_CMD:
	if (protocol == PROTOCOL_1) {
	    close_all_windows();
	    send_to_server(P1_FN_MAINT | P1_MF_EXIT);
	}
	protocol = PROTOCOL_0;
	if (open_window(0, msg.text[REAL_WINDOW], READ_ONLY) < 0)
	    send_controller(P1_INT_DIE);
	else
	    window = 0;
	break;
    case PROTOCOL_1_CMD:
	if (protocol == PROTOCOL_0)
	    close_window(0);
	if (protocol == PROTOCOL_1)
	    close_all_windows();
	protocol = PROTOCOL_1;
	window = -1;
	break;
    case CREATE_WINDOW_CMD:
	if (msg.text[REAL_WINDOW] <= 0) {
	    send_to_server(P1_FN_KILLW | window);
	}
	else {
	    if (open_window(msg.text[LOGICAL_WINDOW], msg.text[REAL_WINDOW],
		READ_ONLY) < 0)
		send_controller(P1_FN_KILLW | msg.text[LOGICAL_WINDOW]);
	    else {
		window = msg.text[LOGICAL_WINDOW];
		send_to_server(P1_FN_ISELW | window);
	    }
	}
	break;
    case KILL_WINDOW_CMD:
	close_window(msg.text[LOGICAL_WINDOW]);
	if (window == LOGICAL_WINDOW)
	    window = -1;
	break;
    case DIE_CMD:
	close(phone_fd);
	exit(0);
	break;
    case PAUSE_CMD:
	while (msgrcv(r_id, &msg, MSG_SIZE, MSG_TYPE, 0) < 0)
	    pause();
	while (msgrcv(r_id, &msg, MSG_SIZE, MSG_TYPE, 0) < 0)
	    pause();
	text_line = window;
	t.ut_num = WTXTSLK1;
	strncpy(t.ut_text, &msg.text[STRING_VALUE], WTXTLEN-1);
	t.ut_text[WTXTLEN-1] = '\0';
	ioctl(win_info[text_line].fd, WIOCSETTEXT, &t);
	break;
    }
}

static void
send_to_server(unsigned char c)
{
    static const unsigned char prefix = P1_IAC;

    write(phone_fd, &prefix, 1);
    c = c | P1_DIR_CTOS;
    write(phone_fd, &c, 1);
}

#define CTL(x)	(x & 0x1f)

static void
prompt_user(int fd, const char *prompt, char input[WTXTLEN])
{
    struct utdata t;
    char *p, buffer[WTXTLEN];
    int size, n;
    char c;

    t.ut_num = WTXTSLK1;

    memset(buffer, '\0', WTXTLEN);
    strncpy(buffer, input, WTXTLEN-1);

    strcpy(t.ut_text, prompt);
    strcat(t.ut_text, ": ");
    p = &t.ut_text[strlen(t.ut_text)];
    t.ut_text[WTXTLEN-1] = '\0';

    size = (WTXTLEN-1) - (p-t.ut_text);
    n = strlen(input);
    if (n > size)
	n = size;

    memset(p, ' ', size);
    strncpy(p, buffer, size);
    ioctl(fd, WIOCSETTEXT, &t);

    while (1) {
	while (read(fd, &c, 1) <= 0)
	    ;
	switch(c) {
	    case CTL('D'):
	    case CTL('J'):
	    case CTL('M'):
		memset(t.ut_text, ' ', WTXTLEN-1);
		t.ut_text[WTXTLEN-1] = '\0';
		ioctl(fd, WIOCSETTEXT, &t);
		strcpy(input, buffer);
		return;
	    case CTL('U'):
	    case CTL('X'):
		memset(buffer, '\0', WTXTLEN);
		n = 0;
		break;
	    case CTL('H'):
	    case '\177':
		if (n > 0)
		    buffer[--n] = '\0';
		break;
	    default:
		if (n < size)
		    buffer[n++] = c;
		break;
	}
	memset(p, ' ', size);
	strncpy(p, buffer, size);
	ioctl(fd, WIOCSETTEXT, &t);
    }
}
