/*
 *	uw -- the main program
 *
 * Copyright 1989 by W J Andrew Fyfe.  All rights reserved.  Permission to
 * copy this program is given provided that the copy is not sold and that
 * this copyright notice is included.
 */
 
#include "uw.h"
#include "config.h"

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/window.h>

char *cmd_name;

static volatile void usage(void);
static volatile void terminate(void);
static void reap(void);
static void watch_lock(void);

int phone_fd;
static char *phone_line = DEFAULT_PHONE_LINE;
struct window_info win_info[MAX_WINDOWS + 1];
int protocol = PROTOCOL_NONE;
int pid_rd, pid_wr, pid_lk;
int use_meta, kbd_esc_char;

static int conv_esc(const char *);

extern char *optarg;
extern int optind;

void
main(int argc, char *argv[])
{
    int i;
    char *baud_rate = DEFAULT_BAUD_RATE;
    char *phone_number = 0;

    cmd_name = argv[0];

    if (ioctl(STDIN, WIOCGCURR, (char *)0) < 0) {
	fprintf(stderr, "%s: can't get current window\n", cmd_name);
	exit(1);
    }
    if (init_window() < 0)
	exit(1);

    for (i=0; i<=MAX_WINDOWS; ++i) {
	win_info[i].window_number = -1;
	win_info[i].fd = -1;
    }


#ifdef DEFAULT_META
    use_meta = 1;
#else
    use_meta = 0;
#endif

#ifdef DEFAULT_KBD_ESC
    kbd_esc_char = conv_esc(DEFAULT_KBD_ESC);
#else
    kbd_esc_char = -1;
#endif

    while((i = getopt(argc, argv, "mMe:El:s:")) != EOF) {
	switch (i) {
	case 'm':
	    use_meta = 1;
	    break;
	case 'M':
	    use_meta = 0;
	    break;
	case 'e':
	    kbd_esc_char = conv_esc(optarg);
	    break;
	case 'E':
	    kbd_esc_char = -1;
	    break;
	case 'l':
	    phone_line = optarg;
	    break;
	case 's':
	    baud_rate = optarg;
	    break;
	case '?':
	    usage();
	    break;
	}
    }
    if (optind == argc - 1)
	phone_number = argv[optind];
    else if (optind < argc)
	usage();

    printf("uw: connecting");
    if (phone_number)
	printf(", dialing %s", phone_number);
    if (use_meta)
	printf(", using meta key");
    if (kbd_esc_char >= 0) {
	printf(", using escape char \"");
	if (kbd_esc_char & 0200)
	    printf("M");
	i = kbd_esc_char & 0177;
	if (i < ' ')
	    printf("^%c", i + '@');
	else if (i == 0177)
	    printf("^?");
	else
	    printf("%c", i);
	printf("\" ('\\%.3o')", kbd_esc_char);
    }
    printf("\n");

    phone_fd = dial_number(phone_line, baud_rate, phone_number);

    if (phone_fd < 0)
	terminate();

    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGHUP, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);

    if (create_message_queues() < 0)
	terminate();

    pid_rd = fork();
    if (pid_rd < 0) {
	fprintf(stderr, "%s: can't fork [rd]: ", cmd_name);
	perror("");
	terminate();
    }

    if (pid_rd == 0)
	reader();

    pid_wr = fork();
    if (pid_wr < 0) {
	fprintf(stderr, "%s: can't fork [wr]: ", cmd_name);
	perror("");
	kill(pid_rd, SIGKILL);
	terminate();
    }

    if (pid_wr == 0)
	writer();

    pid_lk = fork();
    if (pid_lk < 0) {
	fprintf(stderr, "%s: can't fork [lk]: ", cmd_name);
	perror("");
	kill(pid_rd, SIGKILL);
	kill(pid_wr, SIGKILL);
	terminate();
    }

    if (pid_lk == 0)
	watch_lock();

    controller();
    terminate();
}

static volatile void
usage(void)
{
    fprintf(stderr, "%s: Usage: %s [-m | -M | -E | -e esc]"
	" [-l <line>] [-s <speed>] [<telno>]\n",
	cmd_name, cmd_name);
    exit(1);
}

static volatile void
terminate(void)
{
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGHUP, SIG_IGN);

    reap();

    delete_message_queues();

    close_window(0);
    close_all_windows();

    undial_number(phone_line);

    exit(0);
}

static void
catch(void)
{
}

static void
reap(void)
{
    int pid;

    signal(SIGALRM, catch);
    alarm(10);

    if (pid_lk > 0)
	kill(pid_lk, SIGUSR1);

    while (pid_rd > 0 || pid_wr > 0 || pid_lk > 0) {
	pid = wait((int *)0);
	if (pid < 0) {
	    fprintf(stderr, "%s: error from wait: ", cmd_name);
	    perror("");
	    break;
	}
	if (pid == pid_rd)
	    pid_rd = 0;
	else if (pid == pid_wr)
	    pid_wr = 0;
	else if (pid == pid_lk)
	    pid_lk = 0;
    }

    alarm(0);
    signal(SIGALRM, SIG_DFL);

    if (pid_rd > 0)
	kill(pid_rd, SIGKILL);
    if (pid_wr > 0)
	kill(pid_wr, SIGKILL);
    if (pid_lk > 0)
	kill(pid_lk, SIGKILL);
}

static int
conv_esc(const char *p)
{
    const char *q = p;
    int l = strlen(p);
    int meta = 0;

    if (l > 1) {
	if (q[0] == 'M') {
	    meta = 0200;
	    q++;
	    l--;
	}
    }
    if (l == 1) {
	if (q[0] >= ' ' && q[0] < '\177')
	    return q[0] | meta;
    }
    else if (l == 2) {
	if (q[0] == '^') {
	    if (q[1] >= '@' && q[1] <= '_')
		return (q[1] & 0x1f) | meta;
	    else if (q[1] >= 'a' && q[1] <= 'z')
		return (q[1] & 0x1f) | meta;
	    else if (q[1] == '?')
		return '\177' | meta;
	}
    }

    fprintf(stderr, "%s: invalid escape %s\n", cmd_name, p);
    return -1;
}

static void
watch_lock(void)
{
    signal(SIGUSR1, SIG_DFL);
    while (1)
	pause();
    exit(0);
}
