/*
 *	writer -- take input from the host and send it to the windows.
 *
 * Copyright 1989 by W J Andrew Fyfe.  All rights reserved.  Permission to
 * copy this program is given provided that the copy is not sold and that
 * this copyright notice is included.
 */
 
#include "uw.h"
#include <sys/window.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <signal.h>

#define BUF_SIZE	256

static void check_input(void);
static void check_command(void);

static int window = 0, sig_usr1 = 1, sig_alrm;

static struct output_buffer {
    int count;
    unsigned char buf[BUF_SIZE];
} out_buf[MAX_WINDOWS + 1];

static void
catch_usr1(void)
{
    ++sig_usr1;
    signal(SIGUSR1, catch_usr1);
}

static void
catch_alrm(void)
{
    sig_alrm = 1;
    signal(SIGALRM, catch_alrm);
    alarm(TIMEOUT);
}

void
writer(void)
{
    int i;

    signal(SIGUSR1, catch_usr1);
    signal(SIGALRM, SIG_IGN);
    signal(SIGALRM, catch_alrm);
    alarm(TIMEOUT);

    for (i=0; i<=MAX_WINDOWS; ++i)
	out_buf[i].count = 0;

    while (1) {
	if (sig_usr1 || sig_alrm)
	    check_command();
	check_input();
    }
}

static inline void
flush_output(int window)
{
    if (win_info[window].fd >= 0 && out_buf[window].count > 0) {
	write(win_info[window].fd, out_buf[window].buf, out_buf[window].count);
	out_buf[window].count = 0;
    }
}

static void
check_input(void)
{
    static int escape = 0;
    static unsigned char meta_flag = 0;
    int count, i, j;
    unsigned char c;
    unsigned char buf[BUF_SIZE];

    if (protocol == PROTOCOL_NONE)
	return;

    count = read(phone_fd, buf, sizeof(buf));

    for (i=0; i < count; ++i) {
	c = buf[i];

	if (escape) {
	    escape = 0;
	    if ((c & P1_DIR) != P1_DIR_STOC)
		continue;
	    switch (c & P1_FN) {
	    case P1_FN_NEWW:
		send_controller(c);
		continue;
	    case P1_FN_KILLW:
		send_controller(c);
		flush_output(c & P1_WINDOW);
		close_window(c & P1_WINDOW);
		continue;
	    case P1_FN_MAINT:
		send_controller(c);
		for (j=0; j <= MAX_WINDOWS; ++j)
		    flush_output(j);
		switch(c & P1_MF) {
		case P1_MF_EXIT:
		    if (protocol == PROTOCOL_1)
			close_all_windows();
		    protocol = PROTOCOL_0;
		    window = 0;
		    break;
		case P1_MF_ENTRY:
		    if (protocol == PROTOCOL_0)
			close_window(0);
		    if (protocol == PROTOCOL_1)
			close_all_windows();
		    protocol = PROTOCOL_1;
		    window = -1;
		    break;
		}
		continue;
	    case P1_FN_OSELW:
		if (c & P1_WINDOW)
		    window = c & P1_WINDOW;
		continue;
	    case P1_FN_META:
		meta_flag = META_BIT;
		continue;
	    case P1_FN_CTLCH:
		switch(c & P1_CC) {
		case P1_CC_IAC:
		    c = IAC;
		    break;
		case P1_CC_XON:
		    c = XON;
		    break;
		case P1_CC_XOFF:
		    c = XOFF;
		    break;
		}
		break;
	    }
	}
	else if (c == P1_IAC) {
	    escape = 1;
	    continue;
	}

	if (window >= 0 && out_buf[window].count < sizeof(out_buf[window].buf))
	    out_buf[window].buf[out_buf[window].count++] = c | meta_flag;
	meta_flag = 0;
    }

    for (i=0; i <= MAX_WINDOWS; ++i)
	flush_output(i);

    return;
}

static void
check_command(void)
{
    struct message msg;

    sig_alrm = 0;

    if (msgrcv(w_id, &msg, MSG_SIZE, MSG_TYPE, IPC_NOWAIT) < 0)
	return;

    if (sig_usr1 > 0)
	--sig_usr1;

    switch (msg.text[COMMAND]) {
    case PROTOCOL_0_CMD:
	if (protocol == PROTOCOL_1)
	    close_all_windows();
	if (open_window(0, msg.text[REAL_WINDOW], WRITE_ONLY) < 0)
	    send_controller(P1_INT_DIE);
	else {
	    if (protocol == PROTOCOL_1)
		window = 0;
	    protocol = PROTOCOL_0;
	    flush_output(0);
	}
	break;
    case PROTOCOL_1_CMD:
	break;
    case CREATE_WINDOW_CMD:
	if (msg.text[REAL_WINDOW] > 0 &&
	    open_window(msg.text[LOGICAL_WINDOW], msg.text[REAL_WINDOW],
	    WRITE_ONLY) < 0)
	    send_controller(P1_FN_KILLW | msg.text[LOGICAL_WINDOW]);
	else
	    flush_output(msg.text[LOGICAL_WINDOW]);
	break;
    case KILL_WINDOW_CMD:
	flush_output(msg.text[LOGICAL_WINDOW]);
	close_window(msg.text[LOGICAL_WINDOW]);
	break;
    case DIE_CMD:
	close(phone_fd);
	exit(0);
	break;
    case PAUSE_CMD:
	while (msgrcv(w_id, &msg, MSG_SIZE, MSG_TYPE, 0) < 0)
	    pause();
	break;
    }
}
